from typing import List, Optional

from rdflib import Graph, URIRef
from rdflib.namespace import DCTERMS

from fdk_rdf_parser.classes import Reference, SkosCode, SkosConcept


def extractReferences(graph: Graph, subject: URIRef) -> Optional[List[Reference]]:
    referenceProperties = [
        DCTERMS.hasVersion,
        DCTERMS.isVersionOf,
        DCTERMS.isPartOf,
        DCTERMS.hasPart,
        DCTERMS.references,
        DCTERMS.isReferencedBy,
        DCTERMS.replaces,
        DCTERMS.isReplacedBy,
        DCTERMS.requires,
        DCTERMS.isRequiredBy,
        DCTERMS.relation,
        DCTERMS.source,
    ]

    values = []

    for predicate in referenceProperties:
        referenceURI = graph.value(subject, predicate)

        if referenceURI is not None:
            values.append(
                Reference(
                    referenceType=SkosCode(uri=predicate.toPython()),
                    source=SkosConcept(uri=referenceURI.toPython()),
                )
            )

    return values if len(values) > 0 else None



from typing import Any, List, Optional

from rdflib import Graph, Literal, URIRef
from rdflib.namespace import DCTERMS, RDF, SKOS

from fdk_rdf_parser.classes import SkosConcept
from fdk_rdf_parser.rdf_utils import (
    objectValue,
    resourceList,
    valueTranslations,
)


def extractExtraType(graph: Graph, skosConcept: Any) -> Optional[str]:
    extraType = None

    for typeURIRef in resourceList(graph, skosConcept, RDF.type):
        if isinstance(typeURIRef, URIRef) and typeURIRef != SKOS.Concept:
            extraType = typeURIRef.toPython()

    return extraType


def extractSkosConcept(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[List[SkosConcept]]:
    values = []
    for skosConcept in resourceList(graph, subject, predicate):
        skosConceptUri = None
        if isinstance(skosConcept, URIRef) or isinstance(skosConcept, Literal):
            skosConceptUri = skosConcept.toPython()

        sourceValue = objectValue(graph, skosConcept, DCTERMS.source)

        values.append(
            SkosConcept(
                uri=sourceValue if sourceValue is not None else skosConceptUri,
                prefLabel=valueTranslations(graph, skosConcept, SKOS.prefLabel),
                extraType=extractExtraType(graph, skosConcept),
            )
        )

    return values if len(values) > 0 else None



from dataclasses import dataclass
from typing import Any, List, Optional

from .dcat_resource import PartialDcatResource
from .harvest_meta_data import HarvestMetaData


@dataclass
class DataService(PartialDcatResource):

    id: Optional[str] = None
    harvest: Optional[HarvestMetaData] = None

    endpointDescription: Optional[List[str]] = None
    endpointURL: Optional[List[str]] = None
    mediaType: Optional[List[str]] = None
    servesDataset: Optional[List[str]] = None
    conformsTo: Optional[List[str]] = None

    def addValuesFromDcatResource(self: Any, values: PartialDcatResource) -> Any:
        self.identifier = values.identifier
        self.publisher = values.publisher
        self.title = values.title
        self.description = values.description
        self.uri = values.uri
        self.accessRights = values.accessRights
        self.theme = values.theme
        self.keyword = values.keyword
        self.contactPoint = values.contactPoint
        self.type = values.type
        self.issued = values.issued
        self.modified = values.modified
        self.landingPage = values.landingPage
        self.language = values.language



from dataclasses import dataclass
from typing import Dict, Optional

from fdk_rdf_parser.classes import SkosCode, ThemeEU, ThemeLOS
from .reference_data_client import getReferenceData


@dataclass
class ReferenceData:
    provenancestatement: Optional[Dict[str, SkosCode]] = None
    rightsstatement: Optional[Dict[str, SkosCode]] = None
    frequency: Optional[Dict[str, SkosCode]] = None
    linguisticsystem: Optional[Dict[str, SkosCode]] = None
    referencetypes: Optional[Dict[str, SkosCode]] = None
    openlicenses: Optional[Dict[str, SkosCode]] = None
    location: Optional[Dict[str, SkosCode]] = None
    euThemes: Optional[Dict[str, ThemeEU]] = None
    losThemes: Optional[Dict[str, ThemeLOS]] = None


def getAllReferenceData() -> ReferenceData:
    return ReferenceData(
        provenancestatement=getAndMapReferenceCodes("provenancestatement"),
        rightsstatement=getAndMapReferenceCodes("rightsstatement"),
        frequency=getAndMapReferenceCodes("frequency"),
        linguisticsystem=getAndMapReferenceCodes("linguisticsystem"),
        referencetypes=getAndMapReferenceCodes("referencetypes"),
        openlicenses=getAndMapReferenceCodes("openlicenses"),
        location=getAndMapReferenceCodes("location"),
        euThemes=getAndMapThemesEU(),
        losThemes=getAndMapThemesLOS(),
    )


def getAndMapReferenceCodes(endpoint: str) -> Optional[Dict[str, SkosCode]]:
    codes = getReferenceData(f"/codes/{endpoint}")
    if codes is not None:
        return {
            str(code.get("uri")): SkosCode(
                uri=str(code.get("uri")) if code.get("uri") is not None else None,
                code=str(code.get("code")) if code.get("code") is not None else None,
                prefLabel=code.get("prefLabel")
                if code.get("prefLabel") is not None
                else None,
            )
            for code in codes
        }
    else:
        return None


def getAndMapThemesEU() -> Optional[Dict[str, ThemeEU]]:
    mapped = {}
    themes = getReferenceData("/themes")
    if themes is not None:
        for theme in themes:
            euTheme = ThemeEU()
            euTheme.addValuesFromDict(theme)
            mapped[str(theme.get("id"))] = euTheme
    return mapped if len(mapped) > 0 else None


def getAndMapThemesLOS() -> Optional[Dict[str, ThemeLOS]]:
    mapped = {}
    themes = getReferenceData("/los")
    if themes is not None:
        for theme in themes:
            losTheme = ThemeLOS()
            losTheme.addValuesFromDict(theme)
            mapped[str(theme.get("uri"))] = losTheme
    return mapped if len(mapped) > 0 else None



import os

baseUrl = os.getenv(
    "REFERENCE_DATA_BASE_URI",
    "http://staging.fellesdatakatalog.digdir.no/reference-data",
)


def referenceDataUrl(endpoint: str) -> str:
    return f"{baseUrl}{endpoint}"



from rdflib import Graph, URIRef
from rdflib.namespace import DCTERMS

from fdk_rdf_parser.classes import DataService
from fdk_rdf_parser.rdf_utils import (
    dcatURI,
    objectValue,
    valueList,
)
from .dcat_resource import parseDcatResource
from .harvest_meta_data import extractMetaData


def parseDataService(
    dataServicesGraph: Graph, recordURI: URIRef, dataServiceURI: URIRef
) -> DataService:

    dataService = DataService(
        id=objectValue(dataServicesGraph, recordURI, DCTERMS.identifier),
        harvest=extractMetaData(dataServicesGraph, recordURI),
        endpointURL=valueList(
            dataServicesGraph, dataServiceURI, dcatURI("endpointURL")
        ),
        endpointDescription=valueList(
            dataServicesGraph, dataServiceURI, dcatURI("endpointDescription")
        ),
        mediaType=valueList(dataServicesGraph, dataServiceURI, dcatURI("mediaType")),
        conformsTo=valueList(dataServicesGraph, dataServiceURI, dcatURI("conformsTo")),
        servesDataset=valueList(
            dataServicesGraph, dataServiceURI, dcatURI("servesDataset")
        ),
    )

    dataService.addValuesFromDcatResource(
        parseDcatResource(dataServicesGraph, dataServiceURI)
    )
    return dataService



from .ns import (
    admsURI,
    brURI,
    dcatApiURI,
    dcatApNoURI,
    dcatURI,
    dctURI,
    dqvIsoURI,
    dqvURI,
    owlTimeURI,
    provURI,
    rovURI,
    schemaURI,
    vcardURI,
)
from .utils import (
    objectValue,
    resourceList,
    valueList,
    valueTranslations,
)



from typing import List, Optional

from rdflib import Graph, URIRef
from rdflib.namespace import DCTERMS, SKOS

from fdk_rdf_parser.classes import Subject
from fdk_rdf_parser.rdf_utils import (
    objectValue,
    resourceList,
    valueTranslations,
)


def extractSubjects(graph: Graph, datasetRef: URIRef) -> Optional[List[Subject]]:
    datasetSubjects = []
    for subjectRef in resourceList(graph, datasetRef, DCTERMS.subject):
        datasetSubjects.append(
            Subject(
                uri=subjectRef.toPython() if isinstance(subjectRef, URIRef) else None,
                identifier=objectValue(graph, subjectRef, DCTERMS.identifier),
                prefLabel=valueTranslations(graph, subjectRef, SKOS.prefLabel),
                definition=valueTranslations(graph, subjectRef, SKOS.definition),
            )
        )

    return datasetSubjects if len(datasetSubjects) > 0 else None



from dataclasses import dataclass
from datetime import datetime
from typing import Optional


@dataclass
class Temporal:
    uri: Optional[str] = None
    startDate: Optional[datetime] = None
    endDate: Optional[datetime] = None



from .contactpoint import extractContactPoints
from .dataservice import parseDataService
from .dataset import parseDataset
from .dcat_resource import parseDcatResource
from .distribution import extractDistributions
from .harvest_meta_data import extractMetaData
from .publisher import extractPublisher
from .qualified_attribution import extractQualifiedAttributions
from .quality_annotation import extractQualityAnnotation
from .references import extractReferences
from .skos_concept import extractSkosConcept
from .subject import extractSubjects
from .temporal import extractTemporal



from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional


@dataclass
class HarvestMetaData:
    firstHarvested: Optional[datetime] = None
    changed: Optional[List[str]] = None



from typing import Any, Dict, List, Optional

from rdflib import Graph, URIRef


def objectValue(graph: Graph, subject: URIRef, predicate: URIRef) -> Optional[Any]:
    value = graph.value(subject, predicate)
    return value.toPython() if value is not None else None


def valueList(graph: Graph, subject: URIRef, predicate: URIRef) -> Optional[List[Any]]:
    values = []
    for obj in graph.objects(subject, predicate):
        values.append(obj.toPython())
    values.sort()
    return values if len(values) > 0 else None


def valueTranslations(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[Dict[str, str]]:
    values = {}
    for obj in graph.objects(subject, predicate):
        values[obj.language] = obj.toPython()
    return values if len(values) > 0 else None


def resourceList(graph: Graph, subject: URIRef, predicate: URIRef) -> List[Any]:
    values = []
    for obj in graph.objects(subject, predicate):
        values.append(obj)
    values.sort()
    return values



from typing import Optional

from rdflib import Graph, URIRef
from rdflib.namespace import DCTERMS

from fdk_rdf_parser.classes import PublisherId
from fdk_rdf_parser.rdf_utils import objectValue


def extractPublisher(graph: Graph, subject: URIRef) -> Optional[PublisherId]:
    publisherRef = graph.value(subject, DCTERMS.publisher)

    if publisherRef is None:
        return None
    else:
        return PublisherId(
            uri=publisherRef.toPython() if isinstance(publisherRef, URIRef) else None,
            id=objectValue(graph, publisherRef, DCTERMS.identifier),
        )



import logging
from typing import Optional

import requests

from .utils import organizationUrl


def getRdfOrgData(orgnr: Optional[str]) -> Optional[str]:

    try:
        req = requests.get(organizationUrl(orgnr), headers={"Accept": "text/turtle"},)

        return req.text
    except requests.HTTPError as err:
        logging.error(f"Error response from organization-catalogue ({err})")

    return None



from rdflib import Graph, URIRef
from rdflib.namespace import DCTERMS

from fdk_rdf_parser.classes import HarvestMetaData
from fdk_rdf_parser.rdf_utils import objectValue, valueList


def extractMetaData(graph: Graph, subject: URIRef) -> HarvestMetaData:
    return HarvestMetaData(
        firstHarvested=objectValue(graph, subject, DCTERMS.issued),
        changed=valueList(graph, subject, DCTERMS.modified),
    )



import os
import re
from typing import Optional

baseOrgUrl = os.getenv(
    "ORGANIZATION_CATALOGUE_BASE_URI",
    "https://organization-catalogue.staging.fellesdatakatalog.digdir.no",
)


def organizationUrl(orgnr: Optional[str]) -> str:
    return (
        f"{baseOrgUrl}/organizations/{orgnr}"
        if orgnr is not None
        else f"{baseOrgUrl}/organizations"
    )


def organisationNumberFromUri(uri: str) -> Optional[str]:
    match = re.compile(
        "https://data.brreg.no/enhetsregisteret/api/enheter/(\\d{9})$"
    ).match(uri)
    return match.group(1) if match else None



from rdflib import Graph, URIRef
from rdflib.namespace import DCTERMS, FOAF

from fdk_rdf_parser.classes import PartialDataset
from fdk_rdf_parser.rdf_utils import (
    admsURI,
    dcatApNoURI,
    dcatURI,
    dqvIsoURI,
    objectValue,
    provURI,
    valueList,
    valueTranslations,
)
from .dcat_resource import parseDcatResource
from .distribution import extractDistributions
from .harvest_meta_data import extractMetaData
from .qualified_attribution import extractQualifiedAttributions
from .quality_annotation import extractQualityAnnotation
from .references import extractReferences
from .skos_code import extractSkosCode, extractSkosCodeList
from .skos_concept import extractSkosConcept
from .subject import extractSubjects
from .temporal import extractTemporal


def parseDataset(
    datasetsGraph: Graph, recordURI: URIRef, datasetURI: URIRef
) -> PartialDataset:
    qualityAnnotations = extractQualityAnnotation(datasetsGraph, datasetURI)

    dataset = PartialDataset(
        id=objectValue(datasetsGraph, recordURI, DCTERMS.identifier),
        admsIdentifier=valueList(datasetsGraph, datasetURI, admsURI("identifier")),
        harvest=extractMetaData(datasetsGraph, recordURI),
        accessRightsComment=valueList(
            datasetsGraph, datasetURI, dcatApNoURI("accessRightsComment")
        ),
        distribution=extractDistributions(
            datasetsGraph, datasetURI, dcatURI("distribution")
        ),
        sample=extractDistributions(datasetsGraph, datasetURI, admsURI("sample")),
        spatial=extractSkosCodeList(datasetsGraph, datasetURI, DCTERMS.spatial),
        source=objectValue(datasetsGraph, datasetURI, dcatApNoURI("source")),
        objective=valueTranslations(
            datasetsGraph, datasetURI, dcatApNoURI("objective")
        ),
        page=valueList(datasetsGraph, datasetURI, FOAF.page),
        temporal=extractTemporal(datasetsGraph, datasetURI),
        subject=extractSubjects(datasetsGraph, datasetURI),
        provenance=extractSkosCode(datasetsGraph, datasetURI, DCTERMS.provenance),
        accrualPeriodicity=extractSkosCode(
            datasetsGraph, datasetURI, DCTERMS.accrualPeriodicity
        ),
        hasAccuracyAnnotation=qualityAnnotations.get(dqvIsoURI("Accuracy").toPython()),
        hasCompletenessAnnotation=qualityAnnotations.get(
            dqvIsoURI("Completeness").toPython()
        ),
        hasCurrentnessAnnotation=qualityAnnotations.get(
            dqvIsoURI("Currentness").toPython()
        ),
        hasAvailabilityAnnotation=qualityAnnotations.get(
            dqvIsoURI("Availability").toPython()
        ),
        hasRelevanceAnnotation=qualityAnnotations.get(
            dqvIsoURI("Relevance").toPython()
        ),
        legalBasisForRestriction=extractSkosConcept(
            datasetsGraph, datasetURI, dcatApNoURI("legalBasisForRestriction")
        ),
        legalBasisForProcessing=extractSkosConcept(
            datasetsGraph, datasetURI, dcatApNoURI("legalBasisForProcessing")
        ),
        legalBasisForAccess=extractSkosConcept(
            datasetsGraph, datasetURI, dcatApNoURI("legalBasisForAccess")
        ),
        conformsTo=extractSkosConcept(datasetsGraph, datasetURI, DCTERMS.conformsTo),
        informationModel=extractSkosConcept(
            datasetsGraph, datasetURI, dcatApNoURI("informationModel")
        ),
        references=extractReferences(datasetsGraph, datasetURI),
        qualifiedAttributions=extractQualifiedAttributions(
            datasetsGraph, datasetURI, provURI("qualifiedAttribution")
        ),
    )

    dataset.addValuesFromDcatResource(
        parseDcatResource(graph=datasetsGraph, subject=datasetURI)
    )

    return dataset



from typing import Any, List, Optional

from rdflib import BNode, Graph, URIRef
from rdflib.namespace import DCTERMS

from fdk_rdf_parser.classes import Temporal
from fdk_rdf_parser.rdf_utils import (
    dcatURI,
    objectValue,
    owlTimeURI,
    resourceList,
    schemaURI,
)


def extractTemporal(graph: Graph, subject: URIRef) -> Optional[List[Temporal]]:
    values = []
    for temporalResource in resourceList(graph, subject, DCTERMS.temporal):
        temporalURI = None
        if isinstance(temporalResource, URIRef):
            temporalURI = temporalResource.toPython()

        if isinstance(temporalResource, BNode) or temporalURI is not None:
            startValue = objectValue(graph, temporalResource, dcatURI("startDate"))
            endValue = objectValue(graph, temporalResource, dcatURI("endDate"))

            startValue = (
                objectValue(graph, temporalResource, schemaURI("startDate"))
                if startValue is None
                else startValue
            )
            endValue = (
                objectValue(graph, temporalResource, schemaURI("endDate"))
                if endValue is None
                else endValue
            )

            startValue = (
                extractOwlTimeStart(graph, temporalResource)
                if startValue is None
                else startValue
            )
            endValue = (
                extractOwlTimeEnd(graph, temporalResource)
                if endValue is None
                else endValue
            )

            values.append(
                Temporal(uri=temporalURI, startDate=startValue, endDate=endValue)
            )

    return values if len(values) > 0 else None


def extractOwlTimeInstant(graph: Graph, subject: Any) -> Any:
    dateValue = graph.value(subject, owlTimeURI("inXSDDateTime"))
    dateValue = (
        graph.value(subject, owlTimeURI("inXSDDateTimeStamp"))
        if dateValue is None
        else dateValue
    )
    dateValue = (
        graph.value(subject, owlTimeURI("inXSDDate"))
        if dateValue is None
        else dateValue
    )

    if dateValue is not None:
        return dateValue.toPython()
    else:
        return None


def extractOwlTimeStart(graph: Graph, subject: Any) -> Any:
    hasBeginning = graph.value(subject, owlTimeURI("hasBeginning"))
    if hasBeginning is not None:
        return extractOwlTimeInstant(graph, hasBeginning)
    else:
        return None


def extractOwlTimeEnd(graph: Graph, subject: Any) -> Any:
    hasEnd = graph.value(subject, owlTimeURI("hasEnd"))
    if hasEnd is not None:
        return extractOwlTimeInstant(graph, hasEnd)
    else:
        return None



import logging
from typing import Dict, List, Optional

import requests

from .utils import referenceDataUrl


def getReferenceData(endpoint: str) -> Optional[List[Dict]]:

    try:
        response = requests.get(
            referenceDataUrl(endpoint), headers={"Accept": "application/json"}
        )

        response.raise_for_status()

        return response.json()

    except requests.HTTPError as err:
        logging.error(f"Http error response from reference-data ({err})")
    except Exception as err:
        logging.error(f"Error occurred when getting data from reference-data: {err}")

    return None



from typing import List, Optional

from rdflib import BNode, Graph, URIRef

from fdk_rdf_parser.classes import Publisher, QualifiedAttribution
from fdk_rdf_parser.organizations import organisationNumberFromUri
from fdk_rdf_parser.rdf_utils import (
    dcatURI,
    objectValue,
    provURI,
    resourceList,
)


def extractQualifiedAttributions(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[List[QualifiedAttribution]]:
    values = []
    for resource in resourceList(graph, subject, predicate):
        publisherUri = None
        publisherId = None
        role = None

        if isinstance(resource, URIRef) or isinstance(resource, BNode):
            publisherUri = objectValue(graph, resource, provURI("agent"))

            if isinstance(publisherUri, str):
                publisherId = organisationNumberFromUri(publisherUri)
                role = objectValue(graph, resource, dcatURI("hadRole"))

                values.append(
                    QualifiedAttribution(
                        agent=Publisher(uri=publisherUri, id=publisherId), role=role
                    )
                )

    return values if len(values) > 0 else None



from typing import List, Optional

from rdflib import Graph, URIRef
from rdflib.namespace import DCTERMS

from fdk_rdf_parser.classes import DataDistributionService
from fdk_rdf_parser.rdf_utils import (
    dcatApiURI,
    resourceList,
    valueTranslations,
)
from .skos_concept import extractSkosConcept


def extractDataDistributionServices(
    graph: Graph, subject: URIRef
) -> Optional[List[DataDistributionService]]:
    values = []
    for resource in resourceList(graph, subject, dcatApiURI("accessService")):
        resourceUri = None
        if isinstance(resource, URIRef):
            resourceUri = resource.toPython()

        values.append(
            DataDistributionService(
                uri=resourceUri,
                title=valueTranslations(graph, resource, DCTERMS.title),
                description=valueTranslations(graph, resource, DCTERMS.description),
                endpointDescription=extractSkosConcept(
                    graph, resource, dcatApiURI("endpointDescription")
                ),
            )
        )

    return values if len(values) > 0 else None



from typing import Dict

from rdflib import Graph, URIRef
from rdflib.namespace import FOAF, RDF

from .classes import DataService, Dataset, Publisher, PublisherId, QualifiedAttribution
from .organizations import getRdfOrgData, publisherFromFDKOrgCatalog
from .parse_functions import parseDataService, parseDataset
from .rdf_utils import dcatURI, resourceList
from .reference_data import extendDatasetWithReferenceData, getAllReferenceData


def isDcatType(t: URIRef, graph: Graph, topic: URIRef) -> bool:
    for typeURIRef in resourceList(graph, topic, RDF.type):
        if typeURIRef == t:
            return True

    return False


def parseDataServices(dataServiceRDF: str) -> Dict[str, DataService]:
    dataServices: Dict[str, DataService] = {}

    dataServicesGraph = Graph().parse(data=dataServiceRDF, format="turtle")

    for recordURI in dataServicesGraph.subjects(
        predicate=RDF.type, object=dcatURI("CatalogRecord")
    ):
        primaryTopicURI = dataServicesGraph.value(recordURI, FOAF.primaryTopic)

        if primaryTopicURI and isDcatType(
            dcatURI("DataService"), dataServicesGraph, primaryTopicURI
        ):
            dataServices[primaryTopicURI.toPython()] = parseDataService(
                dataServicesGraph, recordURI, primaryTopicURI
            )

    return dataServices


def parseDatasets(rdfData: str) -> Dict[str, Dataset]:
    datasetsGraph = Graph().parse(data=rdfData, format="turtle")
    fdkOrgs = Graph().parse(data=getRdfOrgData(orgnr=None), format="turtle")
    referenceData = getAllReferenceData()

    datasets: Dict[str, Dataset] = {}

    for recordURI in datasetsGraph.subjects(
        predicate=RDF.type, object=dcatURI("CatalogRecord")
    ):
        primaryTopicURI = datasetsGraph.value(recordURI, FOAF.primaryTopic)
        if primaryTopicURI is not None and isDcatType(
            dcatURI("Dataset"), datasetsGraph, primaryTopicURI
        ):
            partialDataset = parseDataset(datasetsGraph, recordURI, primaryTopicURI)

            dataset = Dataset(
                publisher=publisherFromFDKOrgCatalog(partialDataset.publisher, fdkOrgs)
                if partialDataset.publisher is not None
                else None
            )

            dataset.addValuesFromPartial(values=partialDataset)

            dataset = extendDatasetWithReferenceData(dataset, referenceData)
            dataset = extendDatasetWithOrgsData(dataset, fdkOrgs)

            datasets[primaryTopicURI.toPython()] = dataset

    return datasets


def extendDatasetWithOrgsData(dataset: Dataset, organizationsGraph: Graph) -> Dataset:
    if isinstance(dataset.qualifiedAttributions, list):
        dataset.qualifiedAttributions = list(
            map(
                lambda qa: enhanceQualifiedAttributionAgent(qa, organizationsGraph),
                dataset.qualifiedAttributions,
            )
        )

    return dataset


def enhanceQualifiedAttributionAgent(
    qa: QualifiedAttribution, organizationsGraph: Graph
) -> QualifiedAttribution:
    if isinstance(qa.agent, PublisherId):
        publisher = publisherFromFDKOrgCatalog(
            PublisherId(id=qa.agent.id, uri=qa.agent.uri,), organizationsGraph
        )

        if isinstance(publisher, Publisher):
            qa.agent = publisher

    return qa



from typing import List, Optional

from rdflib import Graph, URIRef
from rdflib.namespace import DCTERMS, FOAF

from fdk_rdf_parser.classes import Distribution
from fdk_rdf_parser.rdf_utils import (
    dcatURI,
    dctURI,
    objectValue,
    resourceList,
    valueList,
    valueTranslations,
)
from .data_distribution_service import extractDataDistributionServices
from .skos_concept import extractSkosConcept


def extractDistributions(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[List[Distribution]]:
    values = []
    for resource in resourceList(graph, subject, predicate):
        resourceUri = None
        if isinstance(resource, URIRef):
            resourceUri = resource.toPython()

        values.append(
            Distribution(
                uri=resourceUri,
                title=valueTranslations(graph, resource, DCTERMS.title),
                description=valueTranslations(graph, resource, DCTERMS.description),
                downloadURL=valueList(graph, resource, dcatURI("downloadURL")),
                accessURL=valueList(graph, resource, dcatURI("accessURL")),
                license=extractSkosConcept(graph, resource, DCTERMS.license),
                conformsTo=extractSkosConcept(graph, resource, DCTERMS.conformsTo),
                page=extractSkosConcept(graph, resource, FOAF.page),
                format=valueList(graph, resource, dctURI("format")),
                type=objectValue(graph, resource, DCTERMS.type),
                accessService=extractDataDistributionServices(graph, resource),
            )
        )

    return values if len(values) > 0 else None



from typing import Dict, List, Optional

from rdflib import Graph, URIRef
from rdflib.namespace import DCTERMS

from fdk_rdf_parser.classes import PartialDcatResource, ThemeEU
from fdk_rdf_parser.rdf_utils import (
    dcatURI,
    objectValue,
    valueList,
    valueTranslations,
)
from .contactpoint import extractContactPoints
from .publisher import extractPublisher
from .skos_code import extractSkosCode, extractSkosCodeList


def parseDcatResource(graph: Graph, subject: URIRef) -> PartialDcatResource:
    return PartialDcatResource(
        identifier=valueList(graph, subject, DCTERMS.identifier),
        publisher=extractPublisher(graph, subject),
        title=valueTranslations(graph, subject, DCTERMS.title),
        description=valueTranslations(graph, subject, DCTERMS.description),
        uri=subject.toPython(),
        accessRights=extractSkosCode(graph, subject, DCTERMS.accessRights),
        theme=extractThemes(graph, subject),
        keyword=extractKeyWords(graph, subject),
        contactPoint=extractContactPoints(graph, subject),
        type=objectValue(graph, subject, DCTERMS.type),
        issued=objectValue(graph, subject, DCTERMS.issued),
        modified=objectValue(graph, subject, DCTERMS.modified),
        landingPage=valueList(graph, subject, dcatURI("landingPage")),
        language=extractSkosCodeList(graph, subject, DCTERMS.language),
    )


def extractThemes(graph: Graph, subject: URIRef) -> Optional[List[ThemeEU]]:
    themes = valueList(graph, subject, dcatURI("theme"))
    if themes is not None and len(themes) > 0:
        return list(map(lambda themeURI: ThemeEU(id=themeURI), themes))
    else:
        return None


def extractKeyWords(graph: Graph, subject: URIRef) -> Optional[List[Dict[str, str]]]:
    values = []
    for keyword in graph.objects(subject, dcatURI("keyword")):
        translation = {}
        translation[keyword.language] = keyword.toPython()
        values.append(translation)
    return values if len(values) > 0 else None



from rdflib import Graph, URIRef
from rdflib.namespace import FOAF, SKOS

from fdk_rdf_parser.classes import Publisher, PublisherId
from fdk_rdf_parser.rdf_utils import brURI, objectValue, rovURI, valueTranslations
from .organizations_client import getRdfOrgData
from .utils import organizationUrl


def publisherFromFDKOrgCatalog(publisherId: PublisherId, orgsGraph: Graph) -> Publisher:
    if publisherId.id is None:
        return returnPublisherId(publisherId)
    else:
        orgRef = URIRef(organizationUrl(publisherId.id))

        if (orgRef, None, None) in orgsGraph:
            return parsePublisher(orgsGraph, orgRef)
        else:
            fdkOrg = getRdfOrgData(orgnr=publisherId.id)

            if fdkOrg is not None:
                orgGraph = Graph().parse(data=fdkOrg, format="turtle")
                return parsePublisher(orgGraph, orgRef)
            else:
                return returnPublisherId(publisherId)


def parsePublisher(graph: Graph, orgRef: URIRef) -> Publisher:
    regDataRef = graph.value(orgRef, rovURI("registration"))

    return Publisher(
        id=objectValue(graph, regDataRef, SKOS.notation),
        uri=orgRef.toPython(),
        name=objectValue(graph, orgRef, rovURI("legalName")),
        orgPath=objectValue(graph, orgRef, brURI("orgPath")),
        prefLabel=valueTranslations(graph, orgRef, FOAF.name),
        organisasjonsform=objectValue(graph, orgRef, rovURI("orgType")),
    )


def returnPublisherId(publisherId: PublisherId) -> Publisher:
    return Publisher(uri=publisherId.uri, id=publisherId.id)



from typing import Any, List, Optional

from rdflib import Graph, URIRef

from fdk_rdf_parser.classes import ContactPoint
from fdk_rdf_parser.rdf_utils import dcatURI, objectValue, resourceList, vcardURI


def extractContactPoints(graph: Graph, subject: URIRef) -> Optional[List[ContactPoint]]:
    values = []
    for resource in resourceList(graph, subject, dcatURI("contactPoint")):
        resourceUri = None
        if isinstance(resource, URIRef):
            resourceUri = resource.toPython()

        values.append(
            ContactPoint(
                uri=resourceUri,
                fullname=objectValue(graph, resource, vcardURI("fn")),
                email=extractHasEmail(graph, resource),
                organizationName=objectValue(
                    graph, resource, vcardURI("hasOrganizationName")
                ),
                organizationUnit=objectValue(
                    graph, resource, vcardURI("organization-unit")
                ),
                hasURL=objectValue(graph, resource, vcardURI("hasURL")),
                hasTelephone=extractHasTelephone(graph, resource),
            )
        )

    return values if len(values) > 0 else None


def extractHasTelephone(graph: Graph, subject: Any) -> Optional[Any]:
    telephone: Optional[Any] = objectValue(graph, subject, vcardURI("hasTelephone"))
    if telephone is None:
        return None
    elif "tel:" in telephone:
        return telephone[4:]
    else:
        return telephone


def extractHasEmail(graph: Graph, subject: Any) -> Optional[Any]:
    email: Optional[Any] = objectValue(graph, subject, vcardURI("hasEmail"))
    if email is None:
        return None
    elif "mailto:" in email:
        return email[7:]
    else:
        return email



from rdflib import URIRef


def dcatURI(subString: str) -> URIRef:
    return URIRef(f"http://www.w3.org/ns/dcat#{subString}")


def admsURI(subString: str) -> URIRef:
    return URIRef(f"http://www.w3.org/ns/adms#{subString}")


def dcatApNoURI(subString: str) -> URIRef:
    return URIRef(f"http://difi.no/dcatno#{subString}")


def owlTimeURI(subString: str) -> URIRef:
    return URIRef(f"https://www.w3.org/TR/owl-time/{subString}")


def schemaURI(subString: str) -> URIRef:
    return URIRef(f"http://schema.org/{subString}")


def vcardURI(subString: str) -> URIRef:
    return URIRef(f"http://www.w3.org/2006/vcard/ns#{subString}")


def dctURI(subString: str) -> URIRef:
    return URIRef(f"http://purl.org/dc/terms/{subString}")


def dqvURI(subString: str) -> URIRef:
    return URIRef(f"http://www.w3.org/ns/dqvNS#{subString}")


def dqvIsoURI(subString: str) -> URIRef:
    return URIRef(f"http://iso.org/25012/2008/dataquality/{subString}")


def provURI(subString: str) -> URIRef:
    return URIRef(f"http://www.w3.org/ns/prov#{subString}")


def dcatApiURI(subString: str) -> URIRef:
    return URIRef(f"http://dcat.no/dcatapi/{subString}")


def rovURI(subString: str) -> URIRef:
    return URIRef(f"http://www.w3.org/ns/regorg#{subString}")


def brURI(subString: str) -> URIRef:
    return URIRef(
        f"https://github.com/Informasjonsforvaltning/organization-catalogue/blob/develop/src/main/resources/ontology/organization-catalogue.owl#{subString}"  # noqa: B950
    )



from typing import Dict, List, Optional

from fdk_rdf_parser.classes import (
    Dataset,
    Distribution,
    Reference,
    SkosCode,
    ThemeEU,
    ThemeLOS,
)
from .reference_data import ReferenceData


def extendDatasetWithReferenceData(dataset: Dataset, refData: ReferenceData) -> Dataset:
    dataset.accessRights = extendSkosCode(dataset.accessRights, refData.rightsstatement)
    dataset.provenance = extendSkosCode(dataset.provenance, refData.provenancestatement)
    dataset.accrualPeriodicity = extendSkosCode(
        dataset.accrualPeriodicity, refData.frequency
    )
    dataset.language = extendSkosCodeList(dataset.language, refData.linguisticsystem)
    dataset.spatial = extendSkosCodeList(dataset.spatial, refData.location)
    dataset.distribution = extendDistributions(
        dataset.distribution, refData.openlicenses
    )
    dataset.references = extendReferenceTypes(
        dataset.references, refData.referencetypes
    )

    splitThemes = splitLOSFromEUThemes(dataset.theme, refData.losThemes)
    dataset.losTheme = extendLOSThemes(splitThemes["los"], refData.losThemes)
    dataset.theme = extendEUThemes(splitThemes["eu"], refData.euThemes)

    return dataset


def extendSkosCode(
    skosCode: Optional[SkosCode], references: Optional[Dict[str, SkosCode]]
) -> Optional[SkosCode]:
    if references is not None:
        uri = skosCode.uri if skosCode is not None else None
        if uri is not None:
            refCode = references.get(uri) if references is not None else None
            if refCode is not None:
                return refCode

    return skosCode


def extendSkosCodeList(
    skosCodes: Optional[List[SkosCode]], references: Optional[Dict[str, SkosCode]]
) -> Optional[List[SkosCode]]:
    if skosCodes is None or references is None:
        return skosCodes
    else:
        extendedCodes = []
        for code in skosCodes:
            refCode = references.get(code.uri) if code.uri is not None else None
            if refCode is not None:
                extendedCodes.append(refCode)
            else:
                extendedCodes.append(code)
        return extendedCodes


def extendDistributions(
    distributions: Optional[List[Distribution]],
    openLicenses: Optional[Dict[str, SkosCode]],
) -> Optional[List[Distribution]]:
    if distributions is None or openLicenses is None:
        return distributions
    else:
        extendedDistributions = []
        for dist in distributions:
            if dist.license is not None:
                extendedLicenses = []
                for lic in dist.license:
                    refCode = openLicenses.get(lic.uri) if lic.uri is not None else None
                    if refCode is not None:
                        dist.openLicense = True
                        if lic.prefLabel is None:
                            lic.prefLabel = refCode.prefLabel
                    extendedLicenses.append(lic)
                dist.license = extendedLicenses
            extendedDistributions.append(dist)
        return extendedDistributions


def extendReferenceTypes(
    references: Optional[List[Reference]], referenceTypes: Optional[Dict[str, SkosCode]]
) -> Optional[List[Reference]]:
    if references is None or referenceTypes is None:
        return references
    else:
        extendedReferences = []
        for ref in references:
            ref.referenceType = extendSkosCode(ref.referenceType, referenceTypes)
            extendedReferences.append(ref)
        return extendedReferences


def splitLOSFromEUThemes(
    themes: Optional[List[ThemeEU]], los: Optional[Dict[str, ThemeLOS]],
) -> Dict[str, List[ThemeEU]]:
    splitThemes: Dict[str, List[ThemeEU]] = {}
    splitThemes["los"] = []
    splitThemes["eu"] = []
    if themes is None:
        return splitThemes
    elif los is None:
        splitThemes["eu"] = themes
        return splitThemes
    else:
        for theme in themes:
            if theme.id in los:
                splitThemes["los"].append(theme)
            else:
                splitThemes["eu"].append(theme)
        return splitThemes


def extendLOSThemes(
    themes: List[ThemeEU], los: Optional[Dict[str, ThemeLOS]]
) -> Optional[List[ThemeLOS]]:
    extended = []
    if los is not None:
        for theme in themes:
            extended.append(los[str(theme.id)])
    return extended if len(extended) > 0 else None


def extendEUThemes(
    themes: List[ThemeEU], euThemes: Optional[Dict[str, ThemeEU]]
) -> Optional[List[ThemeEU]]:
    extended = []
    if euThemes is not None:
        for theme in themes:
            euTheme = euThemes.get(theme.id) if theme.id is not None else None
            if euTheme is None:
                extended.append(theme)
            else:
                extended.append(euTheme)
    return extended if len(extended) > 0 else None



from typing import List, Optional

from rdflib import Graph, URIRef

from fdk_rdf_parser.classes import SkosCode
from fdk_rdf_parser.rdf_utils import objectValue, valueList


def extractSkosCode(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[SkosCode]:
    uri = objectValue(graph, subject, predicate)
    return SkosCode(uri=uri) if uri is not None else None


def extractSkosCodeList(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[List[SkosCode]]:
    skosCodes = []
    uriList = valueList(graph, subject, predicate)
    if uriList is None:
        return None
    else:
        for uri in uriList:
            skosCodes.append(SkosCode(uri=uri))
        return skosCodes if len(skosCodes) > 0 else None



from typing import Dict, Optional

from rdflib import Graph, URIRef
from rdflib.namespace import RDF

from fdk_rdf_parser.classes import QualityAnnotation
from fdk_rdf_parser.rdf_utils import (
    dqvURI,
    objectValue,
    provURI,
    resourceList,
    valueTranslations,
)


def extractQualityAnnotation(
    graph: Graph, subject: URIRef
) -> Dict[str, QualityAnnotation]:
    annotations = {}

    for annotation in resourceList(graph, subject, dqvURI("hasQualityAnnotation")):
        inDimensionValue = objectValue(graph, annotation, dqvURI("inDimension"))
        if inDimensionValue is not None and isinstance(inDimensionValue, str):
            annotations[inDimensionValue] = QualityAnnotation(
                inDimension=inDimensionValue, hasBody=extractHasBody(graph, annotation)
            )

    return annotations


def extractHasBody(graph: Graph, subject: URIRef) -> Optional[Dict[str, str]]:
    hasBodyURI = graph.value(subject, provURI("hasBody"))
    if hasBodyURI is not None:
        return valueTranslations(graph, hasBodyURI, RDF.value)
    else:
        return None
